from __future__ import absolute_import

import sys
from concurrent.futures import Future
from threading import Event, Thread

from tornado import ioloop
from tornado import gen

from .exceptions import ThreadNotStartedError

# Python3's concurrent.futures.Future doesn't allow
# setting exc_info... but exc_info works w/o setting explicitly
_FUTURE_HAS_EXC_INFO = hasattr(Future, "set_exception_info")


class ThreadLoop(object):
    """Tornado IOLoop Backed Concurrent Futures.

    Run Tornado Coroutines from Synchronous Python.

    This is made possible by starting the IOLoop in another thread. When
    coroutines are submitted, they are ran against that loop, and their
    responses are bound to Concurrent Futures.

    .. code-block:: python

        from threadloop import ThreadLoop
        from tornado import gen

        @gen.coroutine
        def coroutine(greeting="Goodbye"):
            yield gen.sleep(1)
            raise gen.Return("%s World" % greeting)

        with ThreadLoop() as threadloop:

            future = threadloop.submit(coroutine, "Hello")

            print future.result() # Hello World

    """
    def __init__(self, io_loop=None):

        self._thread = None
        self._ready = Event()
        self._io_loop = io_loop

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()

    def start(self):
        """Start IOLoop in daemonized thread."""
        assert self._thread is None, 'thread already started'

        # configure thread
        self._thread = Thread(target=self._start_io_loop)
        self._thread.daemon = True

        # begin thread and block until ready
        self._thread.start()
        self._ready.wait()

    def _start_io_loop(self):
        """Start IOLoop then set ready threading.Event."""

        def mark_as_ready():
            self._ready.set()

        if not self._io_loop:
            _io_loop = ioloop.IOLoop.current(instance=False)
            if _io_loop:
                self._io_loop = _io_loop
                mark_as_ready()
                return
            self._io_loop = ioloop.IOLoop()

        self._io_loop.add_callback(mark_as_ready)
        self._io_loop.start()

    def is_ready(self):
        """Is thread & ioloop ready.

        :returns bool:
        """
        if not self._thread:
            return False

        if not self._ready.is_set():
            return False

        return True

    def stop(self):
        """Stop IOLoop & close daemonized thread."""
        self._io_loop.add_callback(self._io_loop.stop)
        self._thread.join()

    def submit(self, fn, *args, **kwargs):
        """Submit Tornado Coroutine to IOLoop in daemonized thread.

        :param fn: Tornado Coroutine to execute
        :param args: Args to pass to coroutine
        :param kwargs: Kwargs to pass to coroutine
        :returns concurrent.futures.Future: future result of coroutine
        """
        if not self.is_ready():
            raise ThreadNotStartedError(
                "The thread has not been started yet, "
                "make sure you call start() first"
            )

        future = Future()

        def execute():
            """Executes fn on the IOLoop."""
            try:
                result = gen.maybe_future(fn(*args, **kwargs))
            except Exception:
                # The function we ran didn't return a future and instead raised
                # an exception. Let's pretend that it returned this dummy
                # future with our stack trace.
                f = gen.Future()
                f.set_exc_info(sys.exc_info())
                on_done(f)
            else:
                result.add_done_callback(on_done)

        def on_done(f):
            """Sets tornado.Future results to the concurrent.Future."""

            if not f.exception():
                future.set_result(f.result())
                return

            # if f is a tornado future, then it has exc_info()
            if hasattr(f, 'exc_info'):
                exception, traceback = f.exc_info()[1:]

            # else it's a concurrent.future
            else:
                # python2's concurrent.future has exception_info()
                if hasattr(f, 'exception_info'):
                    exception, traceback = f.exception_info()

                # python3's concurrent.future just has exception()
                else:
                    exception = f.exception()
                    traceback = None

            # python2 needs exc_info set explicitly
            if _FUTURE_HAS_EXC_INFO:
                future.set_exception_info(exception, traceback)
                return

            # python3 just needs the exception, exc_info works fine
            future.set_exception(exception)

        self._io_loop.add_callback(execute)

        return future



from __future__ import absolute_import

import time
from types import TracebackType
from decimal import Decimal

import pytest
from tornado import gen, ioloop
from concurrent.futures import Future, as_completed, TimeoutError

from threadloop import ThreadLoop
from threadloop.exceptions import ThreadNotStartedError


@pytest.fixture(autouse=True)
def clear_io_loop():
    # clear the current IOLoop before all tests
    ioloop.IOLoop.clear_current()


@pytest.yield_fixture
def threadloop():
    with ThreadLoop() as threadloop:
        yield threadloop


class TestException(Exception):
    pass


def test_coroutine_returns_future(threadloop):

    @gen.coroutine
    def coroutine():
        raise gen.Return("Hello World")

    future = threadloop.submit(coroutine)

    assert isinstance(future, Future), "expected a concurrent.futures.Future"

    assert future.result() == "Hello World"


def test_propogates_arguments(threadloop):

    @gen.coroutine
    def coroutine(message, adjective="Shady"):
        raise gen.Return("Hello %s %s" % (adjective, message))

    future = threadloop.submit(coroutine, "World")
    assert future.result() == "Hello Shady World"

    future = threadloop.submit(coroutine, "World", adjective="Cloudy")
    assert future.result() == "Hello Cloudy World"


def test_coroutine_exception_propagates(threadloop):

    @gen.coroutine
    def coroutine():
        raise TestException()

    with pytest.raises(TestException):
        future = threadloop.submit(coroutine)
        future.result()


def test_coroutine_exception_contains_exc_info(threadloop):

    @gen.coroutine
    def coroutine():
        raise TestException('something went wrong')

    with pytest.raises(Exception) as exc_info:
        threadloop.submit(coroutine).result()

    assert 'something went wrong' in str(exc_info.value)
    assert isinstance(exc_info.value, TestException)
    assert isinstance(exc_info.tb, TracebackType)
    assert (
        "raise TestException('something went wrong')"
        in str(exc_info.traceback[-1])
    )


def test_propagate_concurrent_future_exception(threadloop):
    def func():
        future = Future()
        future.set_exception(AttributeError())
        return future

    with pytest.raises(AttributeError):
        threadloop.submit(func).result()


def test_plain_function(threadloop):

    def not_a_coroutine():
        return "Hello World"

    future = threadloop.submit(not_a_coroutine)

    assert isinstance(future, Future), "expected a concurrent.futures.Future"

    assert future.result() == "Hello World"


def test_plain_function_exception_propagates(threadloop):

    def not_a_coroutine():
        raise TestException()

    future = threadloop.submit(not_a_coroutine)

    with pytest.raises(TestException):
        future = threadloop.submit(not_a_coroutine)
        future.result()


def test_plain_function_exception_contains_exc_info(threadloop):

    def not_a_coroutine():
        raise TestException('something went wrong')

    with pytest.raises(Exception) as exc_info:
        threadloop.submit(not_a_coroutine).result()

    assert 'something went wrong' in str(exc_info.value)
    assert isinstance(exc_info.value, TestException)
    assert isinstance(exc_info.tb, TracebackType)
    assert (
        "raise TestException('something went wrong')"
        in str(exc_info.traceback[-1])
    )


def test_use_existing_ioloop():
    io_loop = ioloop.IOLoop.current()
    threadloop = ThreadLoop(io_loop)

    assert threadloop._io_loop is io_loop

    @gen.coroutine
    def coroutine():
        raise gen.Return("Hello World")

    with threadloop:
        future = threadloop.submit(coroutine)
        assert future.result() == "Hello World"


def test_start_must_be_called_before_submit():
    threadloop = ThreadLoop()

    @gen.coroutine
    def coroutine():
        raise gen.Return("Hello World")

    with pytest.raises(ThreadNotStartedError):
        threadloop.submit(coroutine)


def test_submits_coroutines_concurrently(threadloop):

    @gen.coroutine
    def coroutine1():
        yield gen.sleep(.1)
        raise gen.Return('coroutine1')

    @gen.coroutine
    def coroutine2():
        yield gen.sleep(.1)
        raise gen.Return('coroutine2')

    @gen.coroutine
    def coroutine3():
        yield gen.sleep(.1)
        raise gen.Return('coroutine3')

    start = time.time()

    future1 = threadloop.submit(coroutine1)
    future2 = threadloop.submit(coroutine2)
    future3 = threadloop.submit(coroutine3)

    result1 = future1.result()
    result2 = future2.result()
    result3 = future3.result()

    end = time.time() - start

    # round to float with precision of 1, eg 0.3
    took = float(round(Decimal(str(end)), 1))

    # should only take ~100 ms to finish both
    # instead of ~300ms if they were executed serially
    assert took == .1

    assert result1 == 'coroutine1'
    assert result2 == 'coroutine2'
    assert result3 == 'coroutine3'


def test_as_completed(threadloop):

    @gen.coroutine
    def coroutine1():
        yield gen.sleep(.02)
        raise gen.Return('coroutine1')

    @gen.coroutine
    def coroutine2():
        yield gen.sleep(.03)
        raise gen.Return('coroutine2')

    @gen.coroutine
    def coroutine3():
        yield gen.sleep(.01)
        raise gen.Return('coroutine3')

    @gen.coroutine
    def coroutine4():
        yield gen.sleep(.04)
        raise gen.Return('coroutine4')

    futures = []
    futures.append(threadloop.submit(coroutine1))
    futures.append(threadloop.submit(coroutine2))
    futures.append(threadloop.submit(coroutine3))
    futures.append(threadloop.submit(coroutine4))

    i = 0
    for future in as_completed(futures):
        i = i + 1

        # make sure futures finish in the expected order
        if i == 1:
            assert future.result() == "coroutine3"
        elif i == 2:
            assert future.result() == "coroutine1"
        elif i == 3:
            assert future.result() == "coroutine2"
        elif i == 4:
            assert future.result() == "coroutine4"

    assert i == 4, "expected 4 completed futures"


def test_timeout(threadloop):

    @gen.coroutine
    def too_long():
        yield gen.sleep(5)  # 5 sec task
        raise gen.Return('that was too long')

    start = time.time()
    future = threadloop.submit(too_long)

    with pytest.raises(TimeoutError):
        future.result(timeout=.001)

    end = time.time() - start
    took = float(round(Decimal(str(end)), 1))

    assert took <= .002


def test_block_until_thread_is_ready():

    threadloop = ThreadLoop()

    assert not threadloop.is_ready()

    threadloop.start()

    assert threadloop.is_ready()


def test_is_not_ready_when_ready_hasnt_been_sent():

    threadloop = ThreadLoop()
    threadloop._thread = True  # fake the Thread being set

    assert not threadloop.is_ready()


def test_main_io_loop_is_not_changed():
    threadloop = ThreadLoop()
    threadloop.start()

    # The ThreadLoop's IOLoop should not be the 'current' IOLoop in the main
    # thread.
    tl_loop = threadloop.submit(ioloop.IOLoop.current).result()
    assert ioloop.IOLoop.current() is not tl_loop


def test_ioloop_is_not_already_running():
    threadloop = ThreadLoop()
    threadloop.start()

    @gen.coroutine
    def f():
        yield threadloop.submit(gen.sleep, 0.1)

    ioloop.IOLoop.current().run_sync(f)
